/* -*-Prolog-*-

  An experimental prover
	for ((propositional CTL + BDI logic) with mental state axioms)

						by nide@ics.nara-wu.ac.jp

  In this program, only these mental axioms are introduced:
	INTEND(\phi) <=> BEL(INTEND(\phi))
	INTEND(\alpha) => BEL(\alpha)	where \alpha is an O-formula
  To introduce others, we have to add more inference rules.

  Usage:
    ?- prove(Formula).		...Decides provability
    ?- prove_treedisp(Formula).	...Displays proof tree(only simply) if provable

  Formulas:
	true, false, atomic propositions (as an atom in Prolog)
	F \/ G, F /\ G, ~F, F => G, F <=> G, 
	ax F, af F, ag F, ex F, ef F, eg F,
	bel F, des F, int F (where F, G are formulas)
	F au G, F eu G (stand for A(F U G), E(F U G);
		where F and G are formulas)

  operators combine in this order: unary operators, /\, \/, au/eu, =>, <=>

  Runs on SWI-Prolog, B-Prolog, SICStus Prolog.
  (On B-Prolog and SICStus Prolog, you must comment out part of below)
*/

/* Needed Prolog predicates: append, member, select, intersection, union */

/* on B-Prolog, activate this */
/* intersection(A, B, C) :- intersect(A, B, C). */

/* on SICStus-Prolog, activate this */
/*
:- op(700, xfx, \=).
X \= Y :- X \== Y.
append([], L, L).
append([A|L], M, [A|N]) :- append(L, M, N).
member(A, [A|_]).
member(A, [_|L]) :- member(A, L).
select(A, [A|L], L).
select(A, [B|L], [B|M]) :- select(A, L, M).
intersection([], _, []) :- !.
intersection([A|L], M, [A|N]) :- member(A, M), !, intersection(L, M, N).
intersection([_|L], M, N) :- intersection(L, M, N).
union([], L, L).
union([A|L], M, N) :- member(A, M), !, union(L, M, N).
union([A|L], M, [A|N]) :- union(L, M, N).
*/

:-	op(550, xfx, <=>), op(540, xfx, =>),
	op(530, xfx, au), op(530, xfx, eu), op(530, xfx, aw), op(530, xfx, ew),
	op(530, xfx, an), op(530, xfx, en),
	op(530, xfx, awn), op(530, xfx, ewn),
	op(520, yfx, \/), op(510, yfx, /\), op(500, fy, '~'),
	op(500, fy, ax), op(500, fy, ag), op(500, fy, af),
	op(500, fy, ex), op(500, fy, eg), op(500, fy, ef),
	op(500, fy, bel), op(500, fy, des), op(500, fy, int),
	op(500, fy, pbel), op(500, fy, pdes), op(500, fy, pint),
	op(500, fy, desire), op(500, fy, intend),
	true.

exp_ab(P, P) :- atom(P), P \= true, P \= false, !.
exp_ab(F \/ G, F1 \/ G1) :- exp_ab(F, F1), exp_ab(G, G1).
exp_ab(~F, ~F1) :- exp_ab(F, F1).
exp_ab(ax F, ax F1) :- exp_ab(F, F1).
exp_ab(F au G, F1 au G1) :- exp_ab(F, F1), exp_ab(G, G1).
exp_ab(F eu G, F1 eu G1) :- exp_ab(F, F1), exp_ab(G, G1).
exp_ab(bel F, bel F1) :- exp_ab(F, F1).
exp_ab(des F, des F1) :- exp_ab(F, F1).
exp_ab(int F, int F1) :- exp_ab(F, F1).
exp_ab(true, '_' \/ ~'_') :- !.
exp_ab(false, F1) :- exp_ab(~true, F1), !.
exp_ab(F /\ G, F1) :- exp_ab(~(~F \/ ~G), F1).
exp_ab(F => G, F1) :- exp_ab(~F \/ G, F1).
exp_ab(F <=> G, F1) :- exp_ab((F => G) /\ (G => F), F1).
exp_ab(ex F, F1) :- exp_ab(~ax~F, F1).
exp_ab(af F, F1) :- exp_ab(true au F, F1).
exp_ab(ef F, F1) :- exp_ab(true eu F, F1).
exp_ab(ag F, F1) :- exp_ab(~ef~F, F1).
exp_ab(eg F, F1) :- exp_ab(~af~F, F1).
exp_ab(pbel F, F1) :- exp_ab(~bel~F, F1).
exp_ab(pdes F, F1) :- exp_ab(~des~F, F1).
exp_ab(pint F, F1) :- exp_ab(~int~F, F1).
exp_ab(F aw G, F1) :- exp_ab(~(F/\ ~G eu ~F/\ ~G), F1).
exp_ab(F ew G, F1) :- exp_ab((F eu G) \/ eg F, F1).
exp_ab(F an G, F1) :- exp_ab(~G au F/\G, F1).
exp_ab(F en G, F1) :- exp_ab(~G eu F/\G, F1).
exp_ab(F awn G, F1) :- exp_ab(~G aw F/\G, F1).
exp_ab(F ewn G, F1) :- exp_ab(~G ew F/\G, F1).
exp_ab(desire F, F1) :- exp_ab(des F, F1).
exp_ab(intend F, F1) :- exp_ab(int F, F1).

exp_ab_list([F|L], [F1|L1]) :- exp_ab(F, F1), exp_ab_list(L, L1), !.
exp_ab_list([], []).

 /* only for expanded formulas */
o_fmla(F) :- atom(F).
o_fmla(bel _). o_fmla(des _). o_fmla(int _).
o_fmla(F\/G) :- o_fmla(F), o_fmla(G).
o_fmla(~F) :- i_fmla(F).
o_fmla(F eu G) :- o_fmla(F), o_fmla(G).
i_fmla(F) :- atom(F).
i_fmla(bel _). i_fmla(des _). i_fmla(int _).
i_fmla(F\/G) :- i_fmla(F), i_fmla(G).
i_fmla(~F) :- o_fmla(F).
i_fmla(ax F) :- i_fmla(F).
i_fmla(F au G) :- i_fmla(F), i_fmla(G).

classify([F|Atomic_new], Other_new, [F|New], Atomic, Other) :-
	(atom(F); F = ax(_); F = bel(_); F = des(_); F = int(_)),
	classify(Atomic_new, Other_new, New, Atomic, Other), !.
classify(Atomic_new, [F|Other_new], [F|New], Atomic, Other) :-
	classify(Atomic_new, Other_new, New, Atomic, Other), !.
classify(Atomic, Other, [], Atomic, Other).

sel_ax([F|Flist], [ax F|Alist]) :- \+member(ax F, Alist),
	sel_ax(Flist, Alist), !.
sel_ax(Flist, [_|Alist]) :- sel_ax(Flist, Alist), !.
sel_ax([], []).
sel_bel([F|Flist], [bel F|Blist], [bel F|Alist]):- \+member(bel F, Alist),
	sel_bel(Flist, Blist, Alist), !.
sel_bel(Flist, Blist, [_|Alist]) :- sel_bel(Flist, Blist, Alist), !.
sel_bel([], [], []).
sel_des([F|Flist], [des F|Alist]):- \+member(des F, Alist),
	sel_des(Flist, Alist), !.
sel_des(Flist, [_|Alist]) :- sel_des(Flist, Alist), !.
sel_des([], []).
sel_int([F|Flist], [int F|Alist]):- \+member(int F, Alist),
	sel_int(Flist, Alist), !.
sel_int(Flist, [_|Alist]) :- sel_int(Flist, Alist), !.
sel_int([], []).

sel_bi_ab([F|Flist], [bel F|Blist], [bel F|Alist]):- \+member(bel F, Alist),
	sel_bi_ab(Flist, Blist, Alist), !.
sel_bi_ab(Flist, [int F|Blist], [int F|Alist]):- \+member(int F, Alist),
	sel_bi_ab(Flist, Blist, Alist), !.
sel_bi_ab(Flist, Blist, [_|Alist]) :- sel_bi_ab(Flist, Blist, Alist), !.
sel_bi_ab([], [], []).
sel_bi_coL(Flist1, [bel F|Blist], [bel F|Alist]):- \+member(bel F, Alist),
	(i_fmla(F) -> Flist1 = [F|Flist]; Flist1 = Flist), !,
	sel_bi_coL(Flist, Blist, Alist), !.
sel_bi_coL([F|Flist], [int F|Blist], [int F|Alist]):- \+member(int F, Alist),
	sel_bi_coL(Flist, Blist, Alist), !.
sel_bi_coL(Flist, Blist, [_|Alist]) :- sel_bi_coL(Flist, Blist, Alist), !.
sel_bi_coL([], [], []).
sel_bi_coR(Flist1, [bel F|Blist], [bel F|Alist]):- \+member(bel F, Alist),
	(o_fmla(F) -> Flist1 = [F|Flist]; Flist1 = Flist), !,
	sel_bi_coR(Flist, Blist, Alist), !.
sel_bi_coR(Flist, [int F|Blist], [int F|Alist]):- \+member(int F, Alist),
 /* debug 02/08/30; 1st arg [F|Flist] -> Flist */
	sel_bi_coR(Flist, Blist, Alist), !.
sel_bi_coR(Flist, Blist, [_|Alist]) :- sel_bi_coR(Flist, Blist, Alist), !.
sel_bi_coR([], [], []).

seq_eq(seq(left(LA, LO), right(RA, RO)), seq(left(LA1,LO1), right(RA1,RO1))) :-
	multiset_eq(LA, LA1), multiset_eq(LO, LO1),
	multiset_eq(RA, RA1), multiset_eq(RO, RO1).
multiset_eq([E|L], L1) :- select(E, L1, L2), multiset_eq(L, L2), !.
multiset_eq([], []).

seq_member(S, [Seq|_]) :- seq_eq(S, Seq).
seq_member(S, [_|SeqList]) :- seq_member(S, SeqList).


my_intersection(universal_set, L2, L2) :- !.
my_intersection(L1, universal_set, L1) :- !.
my_intersection(L1, L2, L) :- intersection(L1, L2, L), !.

select_unterm((Seq, EvSet), [(Seq1, EvSet)|PosUnterm], PosUnterm) :-
	seq_eq(Seq, Seq1).
select_unterm(SeqEvPair, [SeqEvPair1|PosUnterm], [SeqEvPair1|PosUnterm1]) :-
	select_unterm(SeqEvPair, PosUnterm, PosUnterm1).

append_unterm([(Seq,EvSet1)|PosUnterm1],PosUnterm2, [(Seq,EvSet)|PosUnterm]) :-
	select_unterm((Seq, EvSet2), PosUnterm2, PosUnterm2rest), !,
	intersection(EvSet1, EvSet2, EvSet),
	append_unterm(PosUnterm1, PosUnterm2rest, PosUnterm), !.
append_unterm([SeqEvPair|PosUnterm1], PosUnterm2, [SeqEvPair|PosUnterm]) :-
	append_unterm(PosUnterm1, PosUnterm2, PosUnterm), !.
append_unterm([], PosUnterm, PosUnterm).


loopinfochkAEU(loopinfo(PosTerm1, PosUnterm1, Neg1),
	       loopinfo(PosTerm2, PosUnterm2, Neg2), Seq, Ev,
	       loopinfo(PosTerm, PosUnterm, Neg)) :-
	(PosUnterm1 = [] -> Neg1new = Neg1; union(Neg1, [Ev], Neg1new)),
	(PosUnterm2 = [] -> PosUnterm2new = PosUnterm2;
		loopinfoevadd(PosUnterm2, Ev, PosUnterm2new)),
	loopinfo2chk(loopinfo(PosTerm1, PosUnterm1, Neg1new),
		     loopinfo(PosTerm2, PosUnterm2new, Neg2), Seq,
		     loopinfo(PosTerm, PosUnterm, Neg)), !.

loopinfoevadd([(Seq, EvSet)|PosUnterm], Ev, [(Seq, EvSet1)|PosUnterm1]) :-
	union(EvSet, [Ev], EvSet1),
	loopinfoevadd(PosUnterm, Ev, PosUnterm1), !.
loopinfoevadd([], _, []).

loopinfochk(loopinfo(PosTerm, PosUnterm, Neg), Seq, LoopinfoNew) :-
	select_unterm((Seq, EvSet), PosUnterm, PosUnterm1), !,
	my_intersection(EvSet, PosTerm, PosTerm1),
	 /* now PosTerm1 \= universal_set */
	(PosUnterm1 = [] -> !,
			member(Ev, PosTerm1), \+member(Ev, Neg),
			LoopinfoNew = loopinfo(universal_set, [], []);
		LoopinfoNew = loopinfo(PosTerm1, PosUnterm1, Neg)), !.
loopinfochk(Loopinfo, _, Loopinfo).
loopinfo2chk(Loopinfo1, Loopinfo2, Seq, LoopinfoNew) :-
	loopinfounite(Loopinfo1, Loopinfo2, Loopinfo),
	loopinfochk(Loopinfo, Seq, LoopinfoNew), !.
loopinfounite(loopinfo(PosTerm1, PosUnterm1, Neg1),
	      loopinfo(PosTerm2, PosUnterm2, Neg2),
	      loopinfo(PosTerm, PosUnterm, Neg)) :-
	my_intersection(PosTerm1, PosTerm2, PosTerm),
	append_unterm(PosUnterm1, PosUnterm2, PosUnterm),
	union(Neg1, Neg2, Neg), !.

prove(F, T) :- prove_seq([], [F], T).
prove_seq(L, R, T) :-
	exp_ab_list(L, L1), exp_ab_list(R, R1),
	classify(LA, LO, L1, [], []), classify(RA, RO, R1, [], []), !,
	prove_seq_trace([seq(left(LA, LO), right(RA, RO))],
			loopinfo(_, [], _), T).
prove(F) :- prove_seq([], [F]).
prove_seq(L, R) :- prove_seq(L, R, notree).
prove_treedisp(F) :- prove_seq_treedisp([], [F]).
prove_seq_treedisp(L, R) :- prove_seq(L, R, T), disp_tree(T).

disp_tree(T) :- disp_tree(T, 0).
disp_tree([C|AL], N) :-
	tab(N), seq_write(C), nl, N1 is N+1, !, disp_trees(AL, N1).
disp_trees([], _).
disp_trees([A|AL], N) :- disp_tree(A, N), !, disp_trees(AL, N).
seq_write(seq(left(LA, LO), right(RA, RO))) :-
	append(LA, LO, L), append(RA, RO, R),
	list_write(L), (L = [] -> true; write(' ')),
	write('->'),
	(R = [] -> true; write(' ')), list_write(R), !.
list_write([]) :- !.
list_write([X]) :- write(X), !.
list_write([X|L]) :- write(X), write(', '), list_write(L).

/* prove_seq_trace(X, _, _) :- write('try: '), write(X), nl, fail. */

 /* initial seq. */
prove_seq_trace([S|_], loopinfo(universal_set, [], []), T) :-
	S = seq(left(LA, LO), right(RA, RO)),
	((member(F, LA), member(F, RA)); (member(F, LO), member(F, RO))),
	(var(T) -> T = [S]; true), !.
 /* loop */
prove_seq_trace([S|Rest], Loopinfo, T) :-
	seq_member(S, Rest), !,
	Loopinfo = loopinfo(universal_set, [(S, [])], []),
	(var(T) -> T = [S]; true), !.
	/* fails if PosUnterm is already bound to [] */

 /* \/-L/R */
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(left(LA, ['_'\/ ~'_'|LO]), R),
	(var(T) -> true; T1 = T), !,
	prove_seq_trace([seq(left(LA, LO), R), S|Rest], Loopinfo1, T1),
	loopinfochk(Loopinfo1, S, Loopinfo),
	(var(T) -> T = [S, T1]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(left(LA, [F\/G|LO]), R),
	classify(LA_new1, LO_new1, [F], LA, LO),
	classify(LA_new2, LO_new2, [G], LA, LO),
	(var(T) -> true; TL = T, TR = T), !,
	prove_seq_trace([seq(left(LA_new1,LO_new1),R), S|Rest], Loopinfo1, TL),
	prove_seq_trace([seq(left(LA_new2,LO_new2),R), S|Rest], Loopinfo2, TR),
	loopinfo2chk(Loopinfo1, Loopinfo2, S, Loopinfo),
	(var(T) -> T = [S, TL, TR]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(L, right(RA, [F\/G|RO])),
	classify(RA_new, RO_new, [F, G], RA, RO),
	(var(T) -> true; T1 = T), !,
	prove_seq_trace([seq(L, right(RA_new,RO_new)), S|Rest], Loopinfo1, T1),
	loopinfochk(Loopinfo1, S, Loopinfo),
	(var(T) -> T = [S, T1]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
 /* ~-L/R */
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(left(LA, [~F|LO]), right(RA, RO)),
	classify(RA_new, RO_new, [F], RA, RO),
	(var(T) -> true; T1 = T), !,
	prove_seq_trace([seq(left(LA, LO), right(RA_new, RO_new)), S|Rest],
		Loopinfo1, T1),
	loopinfochk(Loopinfo1, S, Loopinfo),
	(var(T) -> T = [S, T1]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(left(LA, LO), right(RA, [~F|RO])),
	classify(LA_new, LO_new, [F], LA, LO),
	(var(T) -> true; T1 = T), !,
	prove_seq_trace([seq(left(LA_new, LO_new), right(RA, RO)), S|Rest],
		Loopinfo1, T1),
	loopinfochk(Loopinfo1, S, Loopinfo),
	(var(T) -> T = [S, T1]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
 /* au-L/R */
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(left(LA, [F au G|LO]), R),
	classify(LA_new1, LO_new1, [G], LA, LO),
	classify(LA_new2, LO_new2, [F, ax(F au G)], LA, LO),
	(var(T) -> true; TL = T, TR = T), !,
	prove_seq_trace([seq(left(LA_new1,LO_new1),R), S|Rest], Loopinfo1, TL),
	prove_seq_trace([seq(left(LA_new2,LO_new2),R), S|Rest], Loopinfo2, TR),
	loopinfochkAEU(Loopinfo1, Loopinfo2, S, F au G, Loopinfo),
	(var(T) -> T = [S, TL, TR]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(L, right(RA, [F au G|RO])),
	classify(RA_new1, RO_new1, [F, G], RA, RO),
	classify(RA_new2, RO_new2, [G, ax(F au G)], RA, RO),
	(var(T) -> true; TL = T, TR = T), !,
	prove_seq_trace([seq(L,right(RA_new1,RO_new1)),S|Rest], Loopinfo1, TL),
	prove_seq_trace([seq(L,right(RA_new2,RO_new2)),S|Rest], Loopinfo2, TR),
	loopinfo2chk(Loopinfo1, Loopinfo2, S, Loopinfo),
	(var(T) -> T = [S, TL, TR]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
 /* eu-L/R */
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(left(LA, [F eu G|LO]), R),
	classify(LA_new1, LO_new1, [G], LA, LO),
	classify(LA_new2, LO_new2, [F, ~ax~(F eu G)], LA, LO),
	(var(T) -> true; TL = T, TR = T), !,
	prove_seq_trace([seq(left(LA_new1,LO_new1),R), S|Rest], Loopinfo1, TL),
	prove_seq_trace([seq(left(LA_new2,LO_new2),R), S|Rest], Loopinfo2, TR),
	loopinfochkAEU(Loopinfo1, Loopinfo2, S, F eu G, Loopinfo),
	(var(T) -> T = [S, TL, TR]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(L, right(RA, [F eu G|RO])),
	classify(RA_new1, RO_new1, [F, G], RA, RO),
	classify(RA_new2, RO_new2, [G, ~ax~(F eu G)], RA, RO),
	(var(T) -> true; TL = T, TR = T), !,
	prove_seq_trace([seq(L,right(RA_new1,RO_new1)),S|Rest], Loopinfo1, TL),
	prove_seq_trace([seq(L,right(RA_new2,RO_new2)),S|Rest], Loopinfo2, TR),
	loopinfo2chk(Loopinfo1, Loopinfo2, S, Loopinfo),
	(var(T) -> T = [S, TL, TR]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
 /* ax-KD with Weak */
prove_seq_trace([S|Rest], Loopinfo, T) :-
	S = seq(left(LA, []), right(RA, [])), /* no cut */
	sel_ax(LA_axoff, LA), sel_ax(RA_axoff, RA), /* no cut */
	(RA_axoff = [] -> LA_axoff \= [], RA_axoff1 = [];
		member(R1, RA_axoff), RA_axoff1 = [R1]),
	 /* at least either LA_axoff or RA_axoff1 is not empty */
	classify(LA_new, LO_new, LA_axoff, [], []),
	classify(RA_new, RO_new, RA_axoff1, [], []), /* no cut */
	(var(T) -> true; T1 = T), /* no cut */
	prove_seq_trace([seq(left(LA_new, LO_new), right(RA_new, RO_new)),
		S|Rest], Loopinfo1, T1),
	loopinfochk(Loopinfo1, S, Loopinfo),
	(var(T) -> T = [S, T1]; true),
	(Loopinfo = loopinfo(_, [], _) -> !; true).
 /* des-KD with Weak */
prove_seq_trace([S|Rest], loopinfo(universal_set, [], []), T) :-
	S = seq(left(LA, []), right(RA, [])),
	sel_des(LA_desoff, LA), sel_des(RA_desoff, RA),
	(RA_desoff = [] -> LA_desoff \= [], RA_desoff1 = [];
		member(R1, RA_desoff), RA_desoff1 = [R1]),
	classify(LA_new, LO_new, LA_desoff, [], []),
	classify(RA_new, RO_new, RA_desoff1, [], []),
	(var(T) -> true; T1 = T),
	prove_seq_trace([seq(left(LA_new, LO_new), right(RA_new, RO_new)),
		S|Rest], loopinfo(_, [], _), T1),
	(var(T) -> T = [S, T1]; true), !.
 /* int-KD with Weak */
prove_seq_trace([S|Rest], loopinfo(universal_set, [], []), T) :-
	S = seq(left(LA, []), right(RA, [])),
	sel_int(LA_intoff, LA), sel_int(RA_intoff, RA),
	(RA_intoff = [] -> LA_intoff \= [], RA_intoff1 = [];
		member(R1, RA_intoff), RA_intoff1 = [R1]),
	classify(LA_new, LO_new, LA_intoff, [], []),
	classify(RA_new, RO_new, RA_intoff1, [], []),
	(var(T) -> true; T1 = T),
	prove_seq_trace([seq(left(LA_new, LO_new), right(RA_new, RO_new)),
		S|Rest], loopinfo(_, [], _), T1),
	(var(T) -> T = [S, T1]; true), !.

 /* bel-KD45 (& +-bel about int) with Weak */
prove_seq_trace([S|Rest], loopinfo(universal_set, [], []), T) :-
	S = seq(left(LA, []), right(RA, [])),
	sel_bi_ab(LA_beloff, LA_bionly, LA),
	sel_bi_ab(RA_beloff, RA_bionly, RA),
	(RA_beloff = [] -> LA_beloff \= [], RA_beloff1 = [];
		member(R1, RA_beloff), RA_beloff1 = [R1]),
	classify(LA_new, LO_new, LA_beloff, LA_bionly, []),
	classify(RA_new, RO_new, RA_beloff1, RA_bionly, []),
	(var(T) -> true; T1 = T),
	prove_seq_trace([seq(left(LA_new, LO_new), right(RA_new, RO_new)),
		S|Rest], loopinfo(_, [], _), T1),
	(var(T) -> T = [S, T1]; true),
	!. /* proof without loop; no need of alternate */
 /* bel-int compatibility with Weak */
prove_seq_trace([S|Rest], loopinfo(universal_set, [], []), T) :-
	S = seq(left(LA, []), right(RA, [])),
	sel_bi_coL(LA_intoff, LA_intonly, LA),
	sel_bi_coR(RA_intoff, RA_intonly, RA),
	(RA_intoff = [] -> LA_intoff \= [], RA_intoff1 = [];
		member(R1, RA_intoff), RA_intoff1 = [R1]),
	classify(LA_new, LO_new, LA_intoff, LA_intonly, []),
	classify(RA_new, RO_new, RA_intoff1, RA_intonly, []),
	(var(T) -> true; T1 = T),
	prove_seq_trace([seq(left(LA_new, LO_new), right(RA_new, RO_new)),
		S|Rest], loopinfo(_, [], _), T1),
	(var(T) -> T = [S, T1]; true),
	!. /* proof without loop; no need of alternate */

 /* for test */
testx1  :- prove(af q /\ ag p => p au q).
testx2  :- prove(ef q /\ ag p => p eu q).
testx3  :- prove(af q /\ eg p => p eu q).
testx4  :- prove(p /\ ag(p=>ax p) => ag p).
testx5  :- prove(p /\ ag(p=>ex p) => eg p).
testx6  :- prove(p au q => p \/ s au q \/ r).
testx7  :- prove(p eu q => p \/ s eu q \/ r).
testx8  :- \+ prove(af p \/ ef p \/ ag p \/ eg p).
testx9  :- \+ prove(ag p => p au q).
testx10 :- \+ prove(eg p => p eu q).
testx11 :- \+ prove(p => bel(p)), \+ prove(bel(p) => p).
testx12 :- \+ prove(ax bel(p) => bel(ax p)), \+ prove(bel(ax p) => ax bel(p)).
testxe  :- prove(int af p au bel p => int af p au (bel p \/ ~ bel ef p)).
testx :- testx1, testx2, testx3, testx4, testx5, testx6, testx7,
	 testx8, testx9, testx10, testx11, testx12, testxe.

test2  :- prove(ef p <=> true eu p).
test2b :- prove(ag p <=> ~ef~p).
test3  :- prove(af p <=> true au p).
test3b :- prove(eg p <=> ~af~p).
test4  :- prove(ex(p \/ q) <=> ex p \/ ex q).
test5  :- prove(ax p <=> ~ex~p).
test6  :- prove(p eu q <=> q \/ p /\ ex(p eu q)).
test7  :- prove(p au q <=> q \/ p /\ ax(p au q)).
test8  :- prove(ex true /\ ax true).
test9  :- prove(ag(r => (~q /\ (p => ex r))) => (r => ~(p au q))).
test9b :- prove(ag(r => (~q /\ ex r)) => (r => ~af q)).
test10 :- prove(ag(r => (~q /\ (p => ax r))) => (r => ~(p eu q))).
test10b:- prove(ag(r => (~q /\ ax r)) => (r => ~ef q)).
test11 :- prove(ag(p => q) => (ex p => ex q)).
test12 :- prove(ag true).
testb1 :- prove(bel(p) /\ bel(p=>q) => bel(q)), prove(~bel ~true).
testb2 :- prove(bel(p) => bel(bel(p))).
testb3 :- prove(~bel(~p) => bel(~bel(~p))).
testb4 :- \+ prove(bel(p) => bel(q)).
testd1 :- prove(des(p) /\ des(p=>q) => des(q)), prove(~des ~true).
testi1 :- prove(int(p) /\ int(p=>q) => int(q)), prove(~int ~true).
test0 :- test2, test2b, test3, test3b, test4, test5, test6, test7, !,
	 test8, test9, test9b, test10, test10b, test11, test12, !,
	 testb1, testb2, testb3, testb4, testd1, testi1.

for(X) :- X > 0.
for(X) :- X > 1, X1 is X - 1, for(X1).
testtime :- for(100), testtime1, write(a), fail; nl.
testtime1 :- testxe, !.

/* mental state consistency example */
testm :- prove(bel int af p /\ (int af p => int af p au (bel p \/ ~ bel ef p))
	/\ ag(bel ef p /\ ~ bel p => ax bel ef p) => af bel p).
